/*
 * =====================================================================================
 *
 *        Filename:  applet.c
 *
 *     Description:  main of the Gnome VM Applet
 *
 *         Version:  $Id$
 *         Created:  14.12.2005 09:07:36 CET
 *
 *          Author:  Karel Zak <kzak@redhat.com>
 *
 * =====================================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <config.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <gtk/gtk.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/gnome-help.h>
#include <panel-applet.h>
#include <panel-applet-gconf.h>
#include <gconf/gconf-client.h>

#include "applet.h"


static void vm_about_cb(BonoboUIComponent *, VmAppletCtl *, const char * );
static void vm_help_cb(BonoboUIComponent *, VmAppletCtl *, const char * );
static void vm_prop_cb(BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb );

static void vm_ctl_destroy(VmAppletCtl *ctl);
static void vm_destroy_cb(GtkWidget *widget, VmAppletCtl *ctl);

static GtkWidget *vm_mitem_os(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_state(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_activity(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_id(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_name(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_maxmem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_usemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_ncpus(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_cputime(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_uuid(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);

static void vm_popmenu_update(VmAppletCtl *ctl);
static gint vm_update(VmAppletCtl *ctl);
static void vm_icon_update(VmAppletCtl *ctl, int sz);
static void vm_popmenu_enable_separator(VmAppletCtl *ctl);

static int instances;

static const BonoboUIVerb vm_menu_verbs [] =
{
    BONOBO_UI_UNSAFE_VERB ("VmAppletProp",  vm_prop_cb),
    BONOBO_UI_UNSAFE_VERB ("VmAppletHelp",  vm_help_cb),
    BONOBO_UI_UNSAFE_VERB ("VmAppletAbout", vm_about_cb),
    BONOBO_UI_VERB_END
};

/* Applet Icons */
static const VmAppletStockIcon vm_stock_icons[] =
{
    { VM_STOCK_BLOCKED,     VM_ICONDIR "/vm-blocked.png"     },
    { VM_STOCK_PAUSED,      VM_ICONDIR "/vm-paused.png"      },
    { VM_STOCK_RUNNING,     VM_ICONDIR "/vm-running.png"     },
    { VM_STOCK_SHUTDOWN,    VM_ICONDIR "/vm-shutdown.png"    },
    { VM_STOCK_SHUTOFF,     VM_ICONDIR "/vm-shutoff.png"     },
    { VM_STOCK_UNKNOWN,     VM_ICONDIR "/vm-unknown.png"     },
    { VM_STOCK_DO_PAUSE,    VM_ICONDIR "/vm-do-suspend.png"  },
    { VM_STOCK_DO_SHUTDOWN, VM_ICONDIR "/vm-do-shutdown.png" },
    { VM_STOCK_DO_RESUME,   VM_ICONDIR "/vm-do-resume.png"   },
};

#define VM_ICON  (VM_ICONDIR "/vm-applet.png")
#define VM_XICON  (VM_ICONDIR "/vm-xapplet.png")

/* Applet domains menu fields */
static VmAppletMenuDesc vm_menu_desc[] =
{
    /* data Id, title, gconf, callback, enable */
    { VM_DATA_ACTIVITY, N_("Activity icon"),    "mitem_activity", vm_mitem_activity, TRUE  },
    { VM_DATA_ID,       N_("Id"),               "mitem_id",     vm_mitem_id,         FALSE  },
    { VM_DATA_NAME,     N_("Name"),             "mitem_name",   vm_mitem_name,       TRUE  },

/* Xen reports strange numbers for MaxMemory, so we don't
 * present this junk to end-users.  --kzak 03/2006
 *  { VM_DATA_MAXMEM,   N_("Max memory"),       "mitem_maxmem", vm_mitem_maxmem,     FALSE },
 */
    { VM_DATA_USEDMEM,  N_("Used memory"),      "mitem_usemem", vm_mitem_usemem,     TRUE  },
    { VM_DATA_NCPUS,    N_("Number of CPUs"),   "mitem_ncpus",  vm_mitem_ncpus,      FALSE },
    { VM_DATA_OSTYPE,   N_("OS type"),          "mitem_os",     vm_mitem_os,         FALSE },
    { VM_DATA_STATE,    N_("State"),            "mitem_state",  vm_mitem_state,      FALSE },
    { VM_DATA_CPUTIME,  N_("CPU Time"),         "mitem_cputime",vm_mitem_cputime,    FALSE },
    { VM_DATA_UUID,     N_("UUID"),             "mitem_uuid",   vm_mitem_uuid,       FALSE }
};

/*
 * Prints and deallocates GError
 */
void
hmm_error(GError **error)
{
    _F();
    fprintf(stderr, "%s\n", (*error)->message);
    g_error_free (*error);
    *error = NULL;
}

/*
 * Initializes icons stock
 */
static void
vm_init_icons(void)
{
    GtkIconFactory *icon_factory;
    gint i;

    _F();

    icon_factory = gtk_icon_factory_new();

    for (i = 0; i < G_N_ELEMENTS(vm_stock_icons); i++)
    {
        VmAppletStockIcon icon = vm_stock_icons[i];
        GtkIconSource *icon_source = gtk_icon_source_new();
        GtkIconSet *icon_set = gtk_icon_set_new();

        gtk_icon_source_set_filename(icon_source, icon.filename);
        gtk_icon_set_add_source(icon_set, icon_source);
        gtk_icon_factory_add(icon_factory, icon.stock_id, icon_set);
        gtk_icon_source_free(icon_source);
        gtk_icon_set_unref(icon_set);
    }

    gtk_icon_factory_add_default(icon_factory);
    g_object_unref(G_OBJECT(icon_factory));

    gtk_window_set_default_icon_name("vm-applet");
}

/*
 * Common function invoked by the 'Help' context menu item and the 'Help'
 * button in the preferences dialog.
 */
void
vm_show_help( VmAppletCtl *ctl, const char *section )
{
    GError *error = NULL;

    _F();

    gnome_help_display_on_screen( "vm-applet", section,
                                gtk_widget_get_screen( ctl->applet ),
                                &error);
    if( error )
    {
        char *message;

        message = g_strdup_printf( _("There was an error displaying help: %s"),
                                 error->message );
        vm_error_dialog (ctl->applet, message);
        g_error_free( error );
        g_free( message );
    }
}

/*
 * Called when the user selects the 'help' menu item.
 */
static void
vm_help_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    _F();
    vm_show_help( ctl, NULL );
}

/*
 * Called when the user selects the 'about' menu item.
 */
static void
vm_about_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    GdkPixbuf *pixbuf = NULL;
    GError *error = NULL;

    const gchar *authors[] = {
        "Karel Zak <kzak@redhat.com>",
        NULL
    };

    _F();

    pixbuf = gdk_pixbuf_new_from_file(VM_ICON, &error);
    if (error)
            hmm_error(&error);

    gtk_show_about_dialog( NULL,
        "name",                _("VM Applet"),
        "version",             VERSION,
        "copyright",           "\xC2\xA9 2005, 2006 Red Hat Inc. ",
        "comments",            _("A utility for monitoring Virtual Machines."),
        "authors",             authors,
        "translator-credits",  _("translator-credits"),
        "logo",               pixbuf,
        NULL );

    g_object_unref(pixbuf); /* XXX */
}

/*
 * Called when the user selects the 'properties' menu item.
 */
static void
vm_prop_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    _F();
    vmprop_show(ctl);
}

/*
 * Called when the user resizes the gnome panel
 */
static void
vm_size_cb( PanelApplet *applet, GtkAllocation *allocation, VmAppletCtl *ctl)
{
    _F();
    vm_icon_update(ctl, MIN(allocation->width, allocation->height));
}

/*
 * Creates applet control struct
 */
static VmAppletCtl *
vm_ctl_new(PanelApplet *applet)
{
    GSList *urilist, *list;
    GError *error = NULL;
    VmAppletCtl *ctl;
    int i;

    _F();

    /* allocate ctl */
    ctl = g_new0(VmAppletCtl, 1);
    ctl->applet = GTK_WIDGET(applet);
    ctl->uid = getuid();
    ctl->instance = instances;
    panel_applet_add_preferences(applet, "/schemas/apps/vm-applet/prefs", NULL);

    /* delay */
    ctl->delay = panel_applet_gconf_get_int(applet, VM_GCONF_DELAY, &error);
    if (error)
        hmm_error(&error);
    _D("read gconf: %s=%d", VM_GCONF_DELAY, ctl->delay);
    ctl->delay = ctl->delay <= 0 ? 1000 : ctl->delay;

    /* initialize menu description */
    for (i=0; i < G_N_ELEMENTS(vm_menu_desc); i++)
    {
        VmAppletMenuDesc *desc = &vm_menu_desc[i];
        gboolean val;

        desc->ctl = ctl;

        val = panel_applet_gconf_get_bool(applet, desc->gconf, &error);
        _D("read gconf: %s=%s", desc->gconf, val ? "TRUE" : "FALSE");

        if (error)
            hmm_error(&error);
        else
            desc->enable = val;
    }

    ctl->desc = vm_menu_desc;
    ctl->ndescs = G_N_ELEMENTS(vm_menu_desc);

    vm_connections_gconf_read(ctl);

    return ctl;
}

/*
 * Deallocates the applet control struct
 */
static void
vm_ctl_destroy(VmAppletCtl *ctl)
{
    _F();

    if (ctl->conn)
        vm_connections_destroy(ctl);

    g_free(ctl);
}

/*
 * (Re)sets applet timer
 */
void
vm_timer_refresh(VmAppletCtl *ctl)
{
    _F();

    if (ctl->pixtimer)
        g_source_remove(ctl->pixtimer);

    _D("setting delay to: %d (instance: %d)", ctl->delay, ctl->instance);
    ctl->pixtimer = g_timeout_add (ctl->delay, (GSourceFunc) vm_update, ctl);
}


/*
 * Updates the tool tip on the gome-panel
 */
static void
vm_tooltip_update(VmAppletCtl *ctl)
{
    char *tip;

    _F();

    if (ctl->ndoms > 0)
    {
        if (ctl->ndoms > 1)
            tip = g_strdup_printf (_("%d virtual machines"), ctl->ndoms);
        else
            tip = g_strdup_printf (_("%d virtual machines"), ctl->ndoms);
    }
    else if (!ctl->naconns) {
        if (!ctl->nconns)
            tip = _("No connection defined");
        else if (!ctl->neconns)
            tip = _("All connections are disabled.");
        else
            tip = _("Failed to connect to the libvirtd.");
    }
    else
        tip = _("Failed to load information about virtual machines.");

    gtk_widget_set_tooltip_text(ctl->applet, tip);

    if (ctl->ndoms > 0)
        g_free(tip);
}

/*
 * Creates or updates the applet
 */
static gint
vm_update(VmAppletCtl *ctl)
{
    int ndoms = 0;

    _D("\n\n---- loop (pid: %d, instance: %d, delay: %d) ----", getpid(), ctl->instance, ctl->delay);
    _F();

    ndoms = ctl->ndoms;

    if (ctl->naconns < ctl->neconns)
        vm_connections_refresh(ctl);

    vm_domains_update(ctl);

    _D("old doms: %d, new ndoms: %d, conns: %d/%d/%d",
                ndoms, ctl->ndoms, ctl->naconns, ctl->neconns, ctl->nconns);

    if (ndoms != ctl->ndoms) {
        if (ctl->image)
            vm_icon_update(ctl, ctl->image_size);
        vm_popmenu_enable_separator(ctl);
    }

    if (ctl->popmenu && ctl->popmenu_up)
        vm_popmenu_update(ctl);

    vm_tooltip_update(ctl);

    return TRUE;
}

/*
 * Deallocates the applet (the end! ;-)
 */
static void
vm_destroy_cb(GtkWidget *widget, VmAppletCtl *ctl)
{
    _F();

    g_source_remove(ctl->pixtimer);

    if (ctl->popmenu)
    {
        gtk_widget_destroy(ctl->popmenu);
        ctl->popmenu = NULL;
    }
    vm_ctl_destroy(ctl);

    --instances;
}


/*
 * Counts pop menu position
 * -- the following function is adapted from gtkmenuitem.c
 */
static void
vm_position_popmenu (GtkMenu *menu, gint *x, gint *y,
                      gboolean *push_in, gpointer ctl)
{
    GtkWidget *widget = ((VmAppletCtl *) ctl)->applet;
    GdkScreen *screen;
    gint twidth, theight, tx, ty;
    GtkTextDirection direction;
    GdkRectangle monitor;
    gint monitor_num;

    _F();

    g_return_if_fail (menu != NULL);
    g_return_if_fail (x != NULL);
    g_return_if_fail (y != NULL);

    if (push_in) *push_in = FALSE;

    direction = gtk_widget_get_direction (widget);

    twidth = GTK_WIDGET (menu)->requisition.width;
    theight = GTK_WIDGET (menu)->requisition.height;

    screen = gtk_widget_get_screen (GTK_WIDGET (menu));
    monitor_num = gdk_screen_get_monitor_at_window (screen, widget->window);
    if (monitor_num < 0) monitor_num = 0;
    gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);

    if (!gdk_window_get_origin (widget->window, &tx, &ty))
    {
        g_warning ("Menu not on screen");
        return;
    }

    tx += widget->allocation.x;
    ty += widget->allocation.y;

    if (direction == GTK_TEXT_DIR_RTL)
        tx += widget->allocation.width - twidth;

    if ((ty + widget->allocation.height + theight) <= monitor.y + monitor.height)
        ty += widget->allocation.height;
    else if ((ty - theight) >= monitor.y)
        ty -= theight;
    else if (monitor.y + monitor.height - (ty + widget->allocation.height) > ty)
        ty += widget->allocation.height;
    else
        ty -= theight;

    *x = CLAMP (tx, monitor.x, MAX (monitor.x, monitor.x + monitor.width - twidth));
    *y = ty;
    gtk_menu_set_monitor (menu, monitor_num);
}

/* nobody loves zombies... */
void
vm_catch_child(int ignore ATTRIBUTE_UNUSED)
{
    int status;
    _D("SIGCHLD detected.. wait for child termination");
    waitpid(-1, &status, 0);
}

int
vm_mngr_exec(VmAppletDomain *domain, char *action)
{
    char *arg0;
    pid_t pid;

    _F();

    /* argv[0] */
    if ((arg0 = strrchr(VM_VIRTMANAGER, '/')))
        ++arg0;
    else
        arg0 = VM_VIRTMANAGER;

    signal(SIGCHLD, vm_catch_child);

    switch((pid=fork()))
    {
        case -1:    /* error */
            return 0;
        case 0:     /* child */
        if (domain && action)
        {
            _D("exec: %s --connect %s %s %s", arg0, domain->conn->uri_str,
                    action, domain->uuid);

            execl(VM_VIRTMANAGER, arg0, "--connect", domain->conn->uri_str,
                            action, domain->uuid, NULL);
        }
        else if (domain)
        {
            _D("exec: %s --connect %s", arg0, domain->conn->uri_str);
            execl(VM_VIRTMANAGER, arg0, "--connect",
                            domain->conn->uri_str, NULL);
        }
        else if (action)
        {
            _D("exec: %s %s", arg0, action);
            execl(VM_VIRTMANAGER, arg0, action, NULL);
        }
        else
        {
            _D("exec: %s", arg0);
            execl(VM_VIRTMANAGER, arg0, NULL);
        }
        exit(EXIT_FAILURE);

        default: /* parent */
            break;
    }
    return 0;
}

static gboolean
vm_mngr_run_cb(GtkWidget *menuitem, GdkEventButton *event, gpointer data)
{
    _F();
    gtk_menu_popdown(GTK_MENU(((VmAppletCtl *) data)->popmenu));
    vm_mngr_exec(NULL, NULL);
    return TRUE;
}


static gboolean
vm_mngr_new_cb(GtkWidget *menuitem, GdkEventButton *event, gpointer data)
{
    _F();
    gtk_menu_popdown(GTK_MENU(((VmAppletCtl *) data)->popmenu));
    vm_mngr_exec(NULL, "--show-domain-creator");
    return TRUE;
}

static gboolean
vm_connect_cb(GtkWidget *menuitem, GdkEventButton *event, gpointer data)
{
    _F();
    gtk_menu_popdown(GTK_MENU(((VmAppletCtl *) data)->popmenu));
    vmconn_dialog((VmAppletCtl *) data);
    return TRUE;
}

static GtkWidget *
vm_mitem_os(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->ostype);
    else
    {
        item = gtk_label_new(domain->ostype);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}


static GtkWidget *
vm_mitem_state(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), vm_domainstate_to_string(domain->state));
    else
    {
        item = gtk_label_new(vm_domainstate_to_string(domain->state));
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_activity(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *id;
    _F();

    if (!(id = vm_domainstate_to_stockid(domain->activity)))
    {
        if (domain->id==0)      /* dom0 */
            id = vm_domainstate_to_stockid(VIR_DOMAIN_RUNNING);
        else
            id = vm_domainstate_to_stockid(0);
    }

    if (item)
        gtk_image_set_from_stock(GTK_IMAGE(item), id, GTK_ICON_SIZE_MENU);
    else
    {
        item =  gtk_image_new_from_stock (id, GTK_ICON_SIZE_MENU);
        gtk_misc_set_alignment (GTK_MISC(item), 0.5, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_id(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str = "";
    _F();

    if (domain->id != -1)
        str = g_strdup_printf("%d", domain->id);

    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    if (str && *str)
        g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_name(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->name);
    else
    {
        item = gtk_label_new(domain->name);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_uuid(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->uuid);
    else
    {
        item = gtk_label_new(domain->uuid);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_maxmem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%lu MB", (unsigned long)(domain->maxmem / 1024UL));
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_usemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str = "";
    _F();

    if (!domain->inactive)
        str = g_strdup_printf("%lu MB (%2.2f%%)",
                (unsigned long)(domain->usemem / 1024UL),
                domain->usemem * 100 / (double) domain->conn->ninfo.memory);
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    if (*str)
        g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_ncpus(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%d %s", domain->ncpus, domain->ncpus > 0 ? "CPUs" : "CPU");
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_cputime(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str = NULL;
    float cputime = domain->cputime;

    _F();

    if (cputime != 0)
    {
        cputime /= 1000000000;
        str = g_strdup_printf("%.1fs", cputime);
    }
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

/*
 * Creates menu item
 */
static GtkWidget *
vm_popmenuitem_new(VmAppletCtl *ctl, VmAppletDomain *dom, int pos)
{
    int i;
    GtkWidget *box;
    GtkWidget *menuitem;

    _F();

    menuitem = gtk_menu_item_new ();
    gtk_widget_set_sensitive (menuitem, TRUE);
    gtk_menu_shell_insert(GTK_MENU_SHELL(ctl->popmenu), menuitem, pos);

#ifdef ENABLE_CONSOLEHELPER
    g_signal_connect(G_OBJECT(menuitem), "button-release-event", G_CALLBACK(vm_menuitem_cb), dom);
#endif

    box = gtk_hbox_new(FALSE, 12);
    gtk_container_add (GTK_CONTAINER(menuitem), box);

    for(i=0; i < ctl->ndescs; i++)
    {
        VmAppletMenuDesc *desc = &ctl->desc[i];
        GtkWidget *subitem;

        if (!desc->enable)
            continue;

        subitem = desc->builder(dom, desc, NULL);
        if (!subitem)
        {
            _D("popmenu-item-new: %s: no item: %s", dom->name, desc->name);
            subitem = gtk_label_new("");    /* dummy column */
        }
        gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);
        gtk_size_group_add_widget(desc->sizegroup, subitem);
    }
    gtk_widget_show_all (menuitem);

    return menuitem;
}

/*
 * Updates fields in the menu item
 */
static GtkWidget *
vm_popmenuitem_update(VmAppletCtl *ctl, VmAppletDomain *dom, GtkWidget *menuitem)
{
    GtkWidget *box;
    GList *list, *li;
    int i=0;


    _F();

    if (!dom->changed)
        return menuitem;

    /* get Box */
    if (!(list = gtk_container_get_children(GTK_CONTAINER(menuitem))))
        return menuitem;
    li = g_list_first(list);
    box = li->data;
    g_list_free(list);

    /* get Items */
    list = gtk_container_get_children(GTK_CONTAINER(box));
    _D("popmenu-item-update: %s: found: %d subitems", dom->name, list ? g_list_length(list) : 0);

    li = g_list_first(list);
    for(i=0; i < ctl->ndescs; i++)
    {
        VmAppletMenuDesc *desc = &ctl->desc[i];
        GtkWidget *subitem;

        if (!desc->enable)
        {
            _D("popmenu-item-disabled: %s: %s", dom->name, desc->name);
            continue;
        }
        if (li==NULL)
        {
            _D("*** out of list! ***");
            break;
        }
        subitem = li->data;
        li = g_list_next(li);

        if (!(dom->changed & desc->did))
        {
            _D("popmenu-item-nochange: %s: %s", dom->name, desc->name);
            continue;
        }
        _D("popmenu-item-change: %s: %s", dom->name, desc->name);
        desc->builder(dom, desc, subitem);
    }
    return menuitem;
}

/*
 * Updates all items in the menu
 */
static void
vm_popmenu_update(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;
    int pos = 0;

    _F();

    while(conn)
    {
        VmAppletDomain *dom;

        for(dom=conn->dom; dom; dom=dom->next, pos++)
        {
            if (dom->menuitem==NULL)
            {
                dom->menuitem = vm_popmenuitem_new(ctl, dom, pos);
            }
            else if (dom->changed==0)
            {
                _D("popmenu-nochange: %s", dom->name);
                continue;
            }
            else
            {
                _D("popmenu-change: %s", dom->name);
                vm_popmenuitem_update(ctl, dom, dom->menuitem);
            }
            dom->changed = 0;
        }
        conn = conn->next;
    }
}

void
vm_popmenu_destroy_items(VmAppletCtl *ctl)
{
    VmAppletConn *conn = ctl->conn;

    _F();

    while(conn)
    {
        VmAppletDomain *dom;

        for(dom=conn->dom; dom; dom=dom->next)
        {
            if (dom->menuitem)
                gtk_widget_destroy(dom->menuitem);
            dom->menuitem = NULL;
        }
        conn = conn->next;
    }
}

/*
 * Called when the menu pop down
 */
static void
vm_popmenu_deactivate_cb(GtkMenuShell *menushell, VmAppletCtl *ctl)
{
    _F();
    ctl->popmenu_up = 0;
}

/*
 * Don't use native gtk_image_menu_item_new_with_label(), it allocates
 * extra space for menuitem icon.
 */
static GtkWidget *
vm_gtk_image_menu_item_new_with_label(const char *label, const char *stock_id)
{
    GtkWidget *menuitem;
    GtkWidget *subitem;
    GtkWidget *box;

    _F();

    menuitem = gtk_menu_item_new();
    box = gtk_hbox_new(FALSE, 12);

    if ((subitem = gtk_image_new_from_stock(stock_id, GTK_ICON_SIZE_MENU)))
    {
        gtk_misc_set_alignment (GTK_MISC(subitem), 0.5, 0.5);
        gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);
    }
    subitem = gtk_label_new(label);
    gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);

    gtk_container_add (GTK_CONTAINER(menuitem), box);
    gtk_widget_set_sensitive (menuitem, TRUE);

    return menuitem;
}

static void
vm_popmenu_enable_separator(VmAppletCtl *ctl)
{
    GList *list, *li;
    GtkWidget *menuitem;

    _F();

    if (!ctl->popmenu)
        return;

    if (!(list = gtk_container_get_children(GTK_CONTAINER(ctl->popmenu))))
        return;
    menuitem = g_list_first(list)->data;
    g_list_free(list);

    if (ctl->ndoms)
        gtk_widget_show_all(menuitem);
    else
        gtk_widget_hide_all(menuitem);
}

/*
 * Creates the applet pop menu
 */
static void
vm_popmenu_new(VmAppletCtl *ctl)
{
    GtkWidget *menuitem;
    int i;

    _F();

    if (ctl->popmenu)
        return;

    ctl->popmenu = gtk_menu_new ();
    ctl->popmenu_up = 0;

    for(i=0; i < ctl->ndescs; i++)
        ctl->desc[i].sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    /* separator */
    menuitem = gtk_separator_menu_item_new();
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    vm_popmenu_enable_separator(ctl);

    /* VM Manager: run */
    menuitem = vm_gtk_image_menu_item_new_with_label(
                    _("Run Virtual Machine Manager..."), GTK_STOCK_EXECUTE);
    gtk_widget_show_all(menuitem);
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    g_signal_connect(G_OBJECT(menuitem), "button-release-event",
                    G_CALLBACK(vm_mngr_run_cb), ctl);

    /* VM Manager: new domain */
    menuitem = vm_gtk_image_menu_item_new_with_label(
                    _("Create A New Virtual Machine..."), GTK_STOCK_NEW);
    gtk_widget_show_all(menuitem);
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    g_signal_connect(G_OBJECT(menuitem), "button-release-event",
                    G_CALLBACK(vm_mngr_new_cb), ctl);

    /* Connection menu */
    menuitem = vm_gtk_image_menu_item_new_with_label(
                    _("Connections..."), GTK_STOCK_CONNECT);
    gtk_widget_show_all(menuitem);
    gtk_container_add(GTK_CONTAINER(ctl->popmenu), menuitem);
    g_signal_connect(G_OBJECT(menuitem), "button-release-event",
                    G_CALLBACK(vm_connect_cb), ctl);

    g_signal_connect(G_OBJECT(ctl->popmenu), "deactivate",
            G_CALLBACK(vm_popmenu_deactivate_cb), ctl);
}

/*
 * Called when user press the button on the gnome applet
 */
static gboolean
vm_button_cb(   GtkWidget *widget,
                        GdkEventButton *event,
                        VmAppletCtl *ctl)
{
    _F();

    if (event->button != 1)
        return FALSE;

    if (!ctl->popmenu)
        vm_popmenu_new(ctl);

    vm_popmenu_update(ctl);

    ctl->popmenu_up = 1;

    gtk_menu_popup (GTK_MENU (ctl->popmenu), NULL, NULL,
                vm_position_popmenu, ctl,
                event->button, event->time);

    return TRUE;
}

/*
 * Updates the icon on the gnome panel
 * -- called when user resizes the panel
 */
static void
vm_icon_update(VmAppletCtl *ctl, int sz)
{
    GdkPixbuf *pixbuf;
    const char *tp;

    _F();

    tp = ctl->ndoms > 0 ? VM_ICON : VM_XICON;

    if (ctl->image && sz == ctl->image_size && tp == ctl->image_type) {
        _D("no change");
        return;
    }

    _D("resizing from %d to %d", ctl->image_size, sz);
    _D("refresing from %s to %s", ctl->image_type, tp);

    if (ctl->image && tp != ctl->image_type)
    {
        _D("destroy old %s icon",  ctl->image_type);
        gtk_widget_destroy(ctl->image);
        ctl->image = NULL;
    }

    ctl->image_size = sz;
    ctl->image_type = tp;

    if (!ctl->image)
    {
        GError *error = NULL;

        _D("loading new %s icon", tp);
        ctl->pixbuf = gdk_pixbuf_new_from_file(tp, &error);
        if (error)
            hmm_error(&error);

        ctl->image = gtk_image_new();
        gtk_container_add(GTK_CONTAINER(ctl->applet), ctl->image);
    }

    _D("scale %s icon", tp);
    pixbuf = gdk_pixbuf_scale_simple(ctl->pixbuf, sz, sz, GDK_INTERP_BILINEAR);
    gtk_image_set_from_pixbuf(GTK_IMAGE(ctl->image), pixbuf);

    g_object_unref(pixbuf);
    gtk_widget_show(ctl->image);
}

/*
 * Called by the factory to fill in the fields for the applet.
 */
static gboolean
vm_setup (PanelApplet *applet)
{
    VmAppletCtl *ctl;
    AtkObject *atk;

    _F();

    if (instances==1)
        vm_init_icons();

    panel_applet_set_flags(applet, PANEL_APPLET_EXPAND_MINOR);

    ctl = vm_ctl_new(applet);


    panel_applet_setup_menu_from_file ( PANEL_APPLET (applet),
                                        DATADIR,
                                        "GNOME_VmApplet.xml",
                                        NULL,
                                        vm_menu_verbs,
                                        ctl);

    if (panel_applet_get_locked_down (PANEL_APPLET (applet)))
    {
        BonoboUIComponent *popup_component;

        popup_component = panel_applet_get_popup_component (PANEL_APPLET (applet));
        bonobo_ui_component_set_prop (popup_component,
                "/commands/preferences",
                "hidden", "1",
                NULL);
    }

    g_signal_connect(G_OBJECT(ctl->applet), "button-release-event",
            G_CALLBACK(vm_button_cb), ctl);
    g_signal_connect(G_OBJECT(ctl->applet), "destroy",
            G_CALLBACK(vm_destroy_cb), ctl);

    /* Workaround.. /usr/bin/panel-test-applets resizing */
#ifndef DEBUG
    g_signal_connect(G_OBJECT(ctl->applet), "size_allocate",
            G_CALLBACK(vm_size_cb), ctl);
#endif

    atk = gtk_widget_get_accessible (ctl->applet);
    if (GTK_IS_ACCESSIBLE (atk))
    {
        atk_object_set_name (atk, _("VmApplet"));
        atk_object_set_description(atk, _("Monitor virtual machines"));
    }

    gtk_widget_show_all(ctl->applet);

    vm_update(ctl);
    vm_icon_update(ctl, panel_applet_get_size(PANEL_APPLET(ctl->applet)));
    vm_timer_refresh(ctl);

    return TRUE;
}

/*
 * "main()" of the applet
 */
static gboolean
vm_factory(PanelApplet *applet, const gchar *iid, gpointer ptr)
{
    gboolean retval = FALSE;

    instances++;
    _D("---- INITIALIZE instance: %d ----", instances);

    _F();

    if (!strcmp(iid, "OAFIID:GNOME_VmApplet"))
        retval = vm_setup(applet);

    return retval;
}

/*
 * Bonobo initialization
 */
PANEL_APPLET_BONOBO_FACTORY("OAFIID:GNOME_VmApplet_Factory",
                             PANEL_TYPE_APPLET,
                             "vm-applet",
                             "0",
                             (PanelAppletFactoryCallback) vm_factory,
                             NULL)


/*
 * vim: set tabstop=4:
 * vim: set shiftwidth=4:
 * vim: set expandtab:
 */
