/*
 * =====================================================================================
 *
 *        Filename:  applet.h
 *
 *         Version:  $Id$
 *         Created:  15.12.2005 20:22:13 CET
 *
 *          Author:  Karel Zak <kzak@redhat.com>
 *
 * =====================================================================================
 */

#ifndef __APPLET_H__
#define __APPLET_H__

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtksizegroup.h>
#include <gtk/gtkwidget.h>
#include <panel-applet.h>
#include <libxml/uri.h>

#include <libvirt/libvirt.h>

/*#define DEBUG   1*/

#ifdef DEBUG
#define _D(_format, _a...)      fprintf(stderr, _format "\n", ##_a)
#define _F()                    fprintf(stderr, "--> %s()\n", __FUNCTION__)
#else
#define _D(_format, _a...)
#define _F()
#endif

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

#define VM_APPLET_BOX_SPACE 6

/*
 * ATTRIBUTE_UNUSED:
 *
 * Macro to flag conciously unused parameters to functions
 */
#ifdef __GNUC__
#ifdef HAVE_ANSIDECL_H
#include <ansidecl.h>
#endif
#ifndef ATTRIBUTE_UNUSED
#define ATTRIBUTE_UNUSED __attribute__((unused))
#endif
#else
#define ATTRIBUTE_UNUSED
#endif

#ifndef __attribute__
/* This feature is available in gcc versions 2.5 and later.  */
# if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 5)
#  define __attribute__(Spec) /* empty */
# endif
#endif


/* dummy */
typedef struct _VmAppletCtl VmAppletCtl;
typedef struct _VmAppletDomain VmAppletDomain;

/*
 * Name of external wrapper
 */
#define VM_WRAPPER_NAME     "vm_applet_wrapper"

/*
 * Icon stock names
 */
#define VM_STOCK_NONE        NULL

#define VM_STOCK_BLOCKED    "vm-stock-blocked"
#define VM_STOCK_PAUSED     "vm-stock-paused"
#define VM_STOCK_RUNNING    "vm-stock-running"
#define VM_STOCK_SHUTDOWN   "vm-stock-shutdown"
#define VM_STOCK_SHUTOFF    "vm-stock-shutoff"
#define VM_STOCK_UNKNOWN    "vm-stock-unknonw"

/* amenu */
#define VM_STOCK_DO_DETAILS     GTK_STOCK_PROPERTIES
#define VM_STOCK_DO_RUN         GTK_STOCK_MEDIA_PLAY
#define VM_STOCK_DO_PAUSE       "vm-stock-do-suspend"
#define VM_STOCK_DO_RESUME      "vm-stock-do-resume"
#define VM_STOCK_DO_SHUTDOWN    "vm-stock-do-shutdown"
#define VM_STOCK_DO_OPEN        GTK_STOCK_OPEN

/*
 * GConf stuff
 */
#define VM_GCONF_DELAY              "delay"
#define VM_GCONF_ACTIVE_URI         "uri_alist"
#define VM_GCONF_INACTIVE_URI       "uri_ilist"


/*
 * Icon description
 */
typedef struct
{
    gchar    *stock_id;
    gchar    *filename;
} VmAppletStockIcon;

/* the domain without CPU time change for VM_BLOCKED_TIME
 * (milliseconds) is marked as "blocked"
 */
#define VM_BLOCKED_TIME     10

/* misc */
#define VM_UUID_SIZ         37    /* bytes */

typedef struct VmAppletConn
{
    virConnectPtr           ptr;        /* libvirt connection */
    char                    *uri_str;   /* URI */
    xmlURIPtr               uri;
    VmAppletDomain          *dom;       /* Vm domains */
    int                     ndoms;      /* number of Vm domains */
    virNodeInfo             ninfo;      /* this node HW info */
    char                    *name;      /* connection screen name */
    gboolean                enabled;
    struct VmAppletConn     *next;
} VmAppletConn;

/*
 * Domain description
 */
typedef struct _VmAppletDomain
{
    int                     id;         /* domain ID */
    char                    *name;      /* domain Name */
    char                    uuid[VM_UUID_SIZ]; /* UUID */
    char                    *ostype;    /* domain OS Type */
    virDomainState          state;      /* domain State */
    virDomainState          activity;   /* real state by State + CpuTime */
    unsigned long           maxmem;     /* domain Max allocated Memory */
    unsigned long           usemem;     /* domain Used Memory */
    unsigned long long      cputime;    /* domain CPU Time */
    int                     cputime_ct; /* number of loops without CPU Time change */
    int                     ncpus;      /* domain number of CPUs */
    GtkWidget               *menuitem;  /* popup menu item */
    VmAppletCtl             *ctl;       /* global applet control struct */
    int                     changed;    /* bit area with info about changed columns */
    int                     inactive;   /* inactive domain */
    VmAppletConn            *conn;      /* parent */
    VmAppletDomain          *next;      /* next domain */
} _VmAppletDomain;

/*
 * Domain data IDs
 */
#define VM_DATA_ID          (1 << 1)
#define VM_DATA_NAME        (1 << 2)
#define VM_DATA_OSTYPE      (1 << 3)
#define VM_DATA_ACTIVITY    (1 << 4)
#define VM_DATA_MAXMEM      (1 << 5)
#define VM_DATA_USEDMEM     (1 << 6)
#define VM_DATA_NCPUS       (1 << 7)
#define VM_DATA_CPUTIME     (1 << 8)
#define VM_DATA_STATE       (1 << 9)
#define VM_DATA_UUID        (1 << 10)

/*
 * Domain menuitem description
 */
typedef struct VmAppletMenuDesc
{
    int                 did;         /* data source ID */
    char                *name;       /* column name */
    char                *gconf;      /* pointer to gnome config */
    GtkWidget           *(*builder)(
                            VmAppletDomain *,
                            struct VmAppletMenuDesc *,
                            GtkWidget *);
    gboolean            enable;      /* visible? */
    GtkSizeGroup        *sizegroup;  /* GTK size group */
    VmAppletCtl         *ctl;        /* global applet control struct */
} VmAppletMenuDesc;

/*
 * Internal applet control struct
 */
typedef struct _VmAppletCtl
{
    GtkWidget           *applet;     /* applet widget */
    GtkWidget           *image;      /* applet icon */
    GdkPixbuf           *pixbuf;     /* icon pixbuf */
    int                 image_size;  /* last used applet icon size */
    const char          *image_type; /* last used applet icon type */
    GtkWidget           *popmenu;    /* list of domais */
    GtkWidget           *propwin;    /* preferences window */
    GtkWidget           *actmenu;    /* actions menu */
    int                 popmenu_up;  /* is the pop menu UP/DOWN */
    VmAppletMenuDesc    *desc;       /* menu columns descriptions */
    int                 ndescs;      /* num of items in "descs" */
    VmAppletDomain      *actdom;     /* selected domain */
    guint               pixtimer;    /* glib timer */
    int                 delay;       /* update delay */
    uid_t               uid;         /* who runs me */
    int                 instance;    /* ID of applet instance */
    int                 ndoms;       /* number of domains */
    char                *terminal_cmd; /* used defined terminal command */
    gboolean            terminal_enb;/* To be, or not to be? That is the question! */
    VmAppletConn        *conn;       /* List of connections */
    int                 nconns;      /* number of all connections */
    int                 neconns;     /* number of enabled connections */
    int                 naconns;     /* number of active connections */
} _VmAppletCtl;


/*
 * D-BUS
 */
#define VM_DBUS_VIRTMANAGER_SERVICE     "com.redhat.virt.manager"
#define VM_DBUS_VIRTMANAGER_PATH        "/com/redhat/virt/manager"
#define VM_DBUS_VIRTMANAGER_INTERFACE   "com.redhat.virt.manager"

typedef struct VmAppletOption
{
    const char *key;    /* name */
    const char *desc;   /* human readable description */
} VmAppletOption;

extern void vm_show_help( VmAppletCtl *ctl, const char *section );
extern void vm_timer_refresh(VmAppletCtl *ctl);
extern void vm_popmenu_destroy_items(VmAppletCtl *ctl);
extern void vm_error_dialog(GtkWidget *applet, char *msg);
extern void hmm_error(GError **error);
extern void vmprop_show(VmAppletCtl *ctl);

extern int vm_mngr_exec(VmAppletDomain *domain, char *action);

/* gtkutils.c */
GtkWidget *vm_frame_new(GtkWidget *parent, const gchar *title);
GtkWidget *vm_entry_new(gchar *text);
GtkWidget *vm_combobox_new(VmAppletOption *opts, size_t nopts, char *active);
GtkWidget *vm_checkbutton_new(const gchar *title, gboolean active);
GtkWidget *vm_spinbutton_new(int from, int to, int step, int val);

GtkWidget *vm_prefbox_add(GtkWidget *parent, GtkSizeGroup *sgrp,
                        const gchar *text, GtkWidget *widget,
			const gchar *postext);


/* connections.c */
extern void vm_connection_free(VmAppletConn *conn);

extern void vm_connection_enable(VmAppletCtl *ctl, VmAppletConn *conn, gboolean status);
extern VmAppletConn *vm_connection_new(const char *uri);
extern void vm_connection_free(VmAppletConn *conn);
extern int vm_connections_refresh(VmAppletCtl *ctl);
extern void vm_connections_destroy(VmAppletCtl *ctl);
extern VmAppletConn *vm_connection_add(VmAppletCtl *ctl, VmAppletConn *conn);

extern void vm_connections_gconf_read(VmAppletCtl *ctl);
extern void vm_connections_gconf_write(VmAppletCtl *ctl);


/* amenu.c */
extern gboolean vm_menuitem_cb(GtkWidget *menuitem,
                        GdkEventButton *event, gpointer data);

extern void vm_catch_child(int ignore ATTRIBUTE_UNUSED);

/* domains.c */
extern char *vm_domainstate_to_stockid(virDomainState state);
extern void vm_domain_free(VmAppletDomain *dom);
extern char *vm_domainstate_to_string(virDomainState state);
extern VmAppletDomain *vm_domainslist_rem(VmAppletDomain *list, VmAppletDomain *dom);
extern VmAppletDomain *vm_domainslist_add(VmAppletDomain *list, VmAppletDomain *dom);
extern VmAppletDomain *vm_domainslist_lookup(VmAppletDomain *list, char *uuid);
extern void vm_domainslist_free(VmAppletDomain *list);


#endif /* __APPLET_H__ */

/*
 * vim: set tabstop=4:
 * vim: set shiftwidth=4:
 * vim: set expandtab:
 */
