#include <stdio.h>
#include <stdlib.h>
#include <config.h>
#include <string.h>

#include <gtk/gtk.h>
#include <libgnome/libgnome.h>
#include <panel-applet.h>
#include <panel-applet-gconf.h>
#include <gconf/gconf-client.h>

#include "applet.h"

GtkWidget *
vm_frame_new(GtkWidget *parent, const gchar *title)
{
    GtkWidget *vbox, *vbox2, *hbox;
    GtkWidget *label;
    gchar *tmp;

    vbox = gtk_vbox_new (FALSE, VM_APPLET_BOX_SPACE);
    gtk_box_pack_start (GTK_BOX(parent), vbox, TRUE, TRUE, 0);

    tmp = g_strdup_printf ("<b>%s</b>", title);
    label = gtk_label_new (NULL);
    gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
    gtk_label_set_markup (GTK_LABEL(label), tmp);
    g_free (tmp);
    gtk_box_pack_start (GTK_BOX(vbox), label, TRUE, FALSE, 0);

    hbox = gtk_hbox_new (FALSE, VM_APPLET_BOX_SPACE);
    gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    label = gtk_label_new ("    ");
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);

    vbox2 = gtk_vbox_new (FALSE, VM_APPLET_BOX_SPACE);
    gtk_box_pack_start (GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);

    return vbox2;
}

GtkWidget *
vm_entry_new(gchar *text)
{
    GtkWidget *entry;

    entry = gtk_entry_new ();
    if (text)
        gtk_entry_set_text (GTK_ENTRY(entry), text);

    return entry;
}

GtkWidget *
vm_combobox_new(VmAppletOption *opts, size_t nopts, char *active)
{
    GtkWidget *combo;
    GtkCellRenderer *renderer;
    int i, idx = -1;
    GtkTreeIter iter;
    GtkListStore *store;

    store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);

    for(i = 0; i < nopts; i++) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                0, _(opts[i].desc),
                1, opts[i].key,
                -1);
        if (active && opts[i].key && strcmp(opts[i].key, active) == 0)
            idx = i;
    }

    combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(store));

    if (idx != -1)
        gtk_combo_box_set_active(GTK_COMBO_BOX(combo), idx);

    renderer = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo),
                    renderer, "text", 0, NULL);

    return combo;
}


GtkWidget *
vm_checkbutton_new(const gchar *title, gboolean active)
{
    GtkWidget *checkbox;

    checkbox = gtk_check_button_new_with_mnemonic(title);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbox), active);
    return checkbox;
}


GtkWidget *
vm_spinbutton_new(int from, int to, int step, int val)
{
    GtkWidget *spin;

    spin = gtk_spin_button_new_with_range(from, to, step);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin), (gdouble) val);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON(spin), GTK_UPDATE_ALWAYS);

    return spin;
}


GtkWidget *
vm_prefbox_add(GtkWidget *parent, GtkSizeGroup *sgrp,
               const gchar *text, GtkWidget *widget,
	       const gchar *postext)
{
	GtkWidget *hbox;
	GtkWidget *label;

	hbox = gtk_hbox_new(FALSE, 6);
	gtk_box_pack_start(GTK_BOX(parent), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	label = gtk_label_new_with_mnemonic(text);
	if (sgrp)
		gtk_size_group_add_widget(sgrp, label);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_label_set_mnemonic_widget(GTK_LABEL(label), widget);
	gtk_widget_show(label);

	gtk_box_pack_start(GTK_BOX(hbox), widget, TRUE, TRUE, 0);

	if (postext) {
		label = gtk_label_new_with_mnemonic(postext);
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	}

	gtk_widget_show(widget);

	return hbox;
}


/*
 * Pops up an error dialog on the same screen as 'applet' saying 'msg'.
 */
void
vm_error_dialog(GtkWidget *applet, char *msg)
{
    GtkWidget *dialog;

    _F();

    dialog = gtk_message_dialog_new( NULL, 0, GTK_MESSAGE_ERROR,
                                     GTK_BUTTONS_OK, "%s", msg);

    gtk_window_set_screen(GTK_WINDOW (dialog),
                           gtk_widget_get_screen (applet));

    gtk_window_set_title(GTK_WINDOW(dialog), _("VM Applet Error"));

    g_signal_connect_swapped( GTK_OBJECT (dialog), "response",
                              G_CALLBACK (gtk_widget_destroy),
                              GTK_OBJECT (dialog) );
    gtk_widget_show_all( dialog );
}
